/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.net.impl;

import org.xmloperator.lambda2.net.model.Cell;
import org.xmloperator.lambda2.net.model.Decoder;
import org.xmloperator.lambda2.net.model.Port;
import org.xmloperator.lambda2.net.model.Stack;

/**
 * Decoder implementation.
 */
public final class DecoderImpl extends CellImpl implements Decoder {
	private static final boolean ASSERTION = Test.ASSERTION;

	private final int index;
	private PortImpl firstAuxiliaryPort = null;
	private Stack auxiliaryPortList = null;
	private long sequenceCount = 1L;

	public DecoderImpl(int index) {
		this.index = index;
	}

	//////////////////////
	// Cell implementation
	//////////////////////

	public final int getSymbol() {
		return SYMBOL_DECODER;
	}

	public final int getSymbolArity() {
		if (this.firstAuxiliaryPort == null)
			return 0;
		if (this.auxiliaryPortList == null)
			return 1;
		return 1 + this.auxiliaryPortList.size();
	}

	public final boolean isDecoder() {
		return true;
	}

	public Port getPort(int portNumber) {
		switch (portNumber) {
			case 0 :
				return this;
			case 1 :
				if (this.firstAuxiliaryPort != null)
					return this.firstAuxiliaryPort;
				break;
			default :
				if (this.auxiliaryPortList != null
						&& portNumber <= this.auxiliaryPortList.size() + 1)
					return (Port) this.auxiliaryPortList.get(portNumber - 2);
		}
		throw new IllegalArgumentException();
	}

	public void collectPorts(Stack portStack) {
		super.collectPorts(portStack);
		if (this.firstAuxiliaryPort != null) {
			portStack.push(this.firstAuxiliaryPort);
			if (this.auxiliaryPortList != null) {
				int portCount = this.auxiliaryPortList.size();
				for (int k = 0; k < portCount; k++)
					portStack.push(this.auxiliaryPortList.get(k));
			}
		}
	}

	public Cell cloneCell() {
		DecoderImpl cell = new DecoderImpl(this.index);
		cell.sequenceCount = this.sequenceCount;
		if (this.firstAuxiliaryPort != null) {
			cell.firstAuxiliaryPort = new PortImpl(cell, 1);
			if (this.auxiliaryPortList != null) {
				cell.auxiliaryPortList = new StackImpl();
				int portCount = this.auxiliaryPortList.size();
				for (int k = 1; k <= portCount; k++)
					cell.auxiliaryPortList.push(new PortImpl(cell, 1 + k));
			}
		}
		return cell;
	}

	public final int interact() {
		PortImpl thisDecoderConnectedPort = this.connectedPort;
		if (thisDecoderConnectedPort.portNumber != 0)
			throw new IllegalStateException("Not an active pair");
		CellImpl cell = thisDecoderConnectedPort.cell;
		switch (cell.getSymbol()) {
			case SYMBOL_WIRE :
				return cell.interact();
			case SYMBOL_ERASER : {
				if (this.firstAuxiliaryPort != null) {
					this.firstAuxiliaryPort.connectedPort.connectTo(cell);
					if (this.auxiliaryPortList != null) {
						int portCount = this.auxiliaryPortList.size();
						for (int k = 0; k < portCount; k++)
							((PortImpl) this.auxiliaryPortList.get(k)).connectedPort
									.connectTo(new EraserCellImpl());
					}
				}
				return INTERACTION_ERASING_DECODER;
			}
			case SYMBOL_DUPLICATOR :
			case SYMBOL_CONSTRUCTOR : {
				BinaryCellImpl binaryCell = (BinaryCellImpl) cell;
				if (!binaryCell.isDuplicator)
					throw new IllegalStateException(
							"Unsupported interaction between a constructor and a decoder");
				PortImpl port1 = binaryCell.port1;
				PortImpl port2 = binaryCell.port2;
				PortImpl connectedPort1 = port1.connectedPort;
				PortImpl connectedPort2 = port2.connectedPort;
				Cell cell11 = connectedPort1.cell;
				Cell cell12 = connectedPort2.cell;
				if (ASSERTION && cell11 == binaryCell || cell11 == this
						|| cell12 == this)
					throw new IllegalStateException("Illegal interconnection");
				boolean isThisSequencable = this.firstAuxiliaryPort != null
						&& this.auxiliaryPortList == null;
				boolean isCell11Sequencable = isThisSequencable
						&& connectedPort1.portNumber == 1 && cell11.isDecoder()
						&& cell11.getSymbolArity() == 1
						&& ((DecoderImpl) cell11).index == this.index;
				boolean isCell12Sequencable = isThisSequencable
						&& connectedPort2.portNumber == 1 && cell12.isDecoder()
						&& cell12.getSymbolArity() == 1
						&& ((DecoderImpl) cell12).index == this.index;
				if (isCell11Sequencable || isCell12Sequencable) {
					binaryCell.connectTo(this.firstAuxiliaryPort.connectedPort);
					if (isCell11Sequencable) {
						((DecoderImpl) cell11).sequenceCount += this.sequenceCount;
						((DecoderImpl) cell11).sequenceCount &= Long.MAX_VALUE;
					} else {
						this.connectTo(connectedPort1);
						this.firstAuxiliaryPort.connectTo(port1);
					}
					if (isCell12Sequencable) {
						((DecoderImpl) cell12).sequenceCount += this.sequenceCount;
						((DecoderImpl) cell12).sequenceCount &= Long.MAX_VALUE;
					} else {
						this.connectTo(connectedPort2);
						this.firstAuxiliaryPort.connectTo(port2);
					}
					return isCell11Sequencable && isCell12Sequencable
							? INTERACTION_MERGING_DECODERS
							: INTERACTION_MERGING_DECODER;
				}
				DecoderImpl decoder2 = (DecoderImpl) this.cloneCell();
				connectedPort1.connectTo(this);
				connectedPort2.connectTo(decoder2);
				if (this.firstAuxiliaryPort != null) {
					{
						PortImpl thisDecoderPort1 = this.firstAuxiliaryPort;
						PortImpl decoder2Port1 = decoder2.firstAuxiliaryPort;
						thisDecoderConnectedPort = thisDecoderPort1.connectedPort;
						if (thisDecoderConnectedPort.cell == this)
							throw new IllegalStateException(
									"Illegal interconnection");
						binaryCell.connectTo(thisDecoderConnectedPort);
						port1.connectTo(thisDecoderPort1);
						port2.connectTo(decoder2Port1);
					}
					if (this.auxiliaryPortList != null) {
						int portCount = this.auxiliaryPortList.size();
						for (int k = 0; k < portCount; k++) {
							BinaryCellImpl duplicator = new BinaryCellImpl(true);
							PortImpl thisDecoderPort = (PortImpl) this.auxiliaryPortList
									.get(k);
							PortImpl decoder2Port = (PortImpl) decoder2.auxiliaryPortList
									.get(k);
							thisDecoderConnectedPort = thisDecoderPort.connectedPort;
							if (thisDecoderConnectedPort.cell == this)
								throw new IllegalStateException(
										"Illegal interconnection");
							duplicator.connectTo(thisDecoderConnectedPort);
							duplicator.port1.connectTo(thisDecoderPort);
							duplicator.port2.connectTo(decoder2Port);
						}
					}
				}
				return INTERACTION_DUPLICATING_DECODER;
			}
			case SYMBOL_ABSTRACTOR :
			case SYMBOL_APPLICATOR : {
				PentaryMultiplexorImpl multiplexor = (PentaryMultiplexorImpl) cell;
				if (multiplexor.isAbstractor)
					throw new IllegalStateException(
							"Unsupported interaction between an abstractor and a decoder");
				if (this.sequenceCount != 1)
					throw new IllegalStateException(
							"Unsupported interaction between an applicator and a sequence of decoders");
				PortImpl connectedPort1 = multiplexor.bodyPort.connectedPort;
				PortImpl connectedPort2 = multiplexor.variablePort.connectedPort;
				PortImpl connectedPort3 = multiplexor.duplicatorPort2.connectedPort;
				PortImpl connectedPort4 = multiplexor.duplicatorPort1.connectedPort;
				PortImpl connectedPort5 = multiplexor.duplicatorPort0.connectedPort;
				Cell cell11 = connectedPort1.cell;
				if (ASSERTION) {
					Cell cell12 = connectedPort2.cell;
					Cell cell13 = connectedPort3.cell;
					Cell cell14 = connectedPort4.cell;
					Cell cell15 = connectedPort5.cell;
					if (cell11 == multiplexor || cell11 == this
							|| cell12 == multiplexor || cell12 == this
							|| cell13 == multiplexor || cell13 == this
							|| cell14 == multiplexor || cell14 == this
							|| cell15 == multiplexor || cell15 == this)
						throw new IllegalStateException(
								"Illegal interconnection");
				}
				if (false && this.firstAuxiliaryPort == null
						&& connectedPort1.portNumber == 1 && cell11.isDecoder()
						&& cell11.getSymbolArity() == 1
						&& ((DecoderImpl) cell11).index == this.index) {
					connectedPort1.connectTo(connectedPort2);
					((DecoderImpl) cell11).sequenceCount++;
					((DecoderImpl) cell11).sequenceCount &= Long.MAX_VALUE;
				} else {
					connectedPort1.connectTo(this);
					connectedPort2.connectTo(this.addNewAuxiliaryPort());
				}
				connectedPort3.connectTo(new EraserCellImpl());
				connectedPort4.connectTo(new EraserCellImpl());
				connectedPort5.connectTo(new EraserCellImpl());
				return INTERACTION_APPLICATOR_DECODER;
			}
			case SYMBOL_DECODER :
				throw new IllegalStateException(
						"Unsupported interaction between two decoders");
			default :
				throw new IllegalStateException("Unknown symbol");
		}
	}

	/////////////////////////
	// Decoder implementation
	/////////////////////////

	public final int getIndex() {
		return this.index;
	}

	public final Port addNewAuxiliaryPort() {
		if (this.firstAuxiliaryPort == null) {
			this.firstAuxiliaryPort = new PortImpl(this, 1);
			return this.firstAuxiliaryPort;
		}
		if (this.auxiliaryPortList == null)
			this.auxiliaryPortList = new StackImpl();
		else if (this.sequenceCount != 1)
			throw new IllegalStateException();
		int portCount = this.auxiliaryPortList.size();
		PortImpl port = new PortImpl(this, portCount + 1);
		this.auxiliaryPortList.push(port);
		return port;
	}

	public long getSequenceCount() {
		return this.sequenceCount;
	}

	public void addSequenceCount(long sequenceCount) {
		if (this.firstAuxiliaryPort == null || this.auxiliaryPortList != null)
			throw new IllegalStateException();
		if (sequenceCount <= 0)
			throw new IllegalArgumentException();
		this.sequenceCount += sequenceCount;
		this.sequenceCount &= Long.MAX_VALUE;
	}
}