/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.net.impl;

import java.util.HashSet;
import java.util.Set;

import org.xmloperator.lambda2.net.model.BinaryCell;
import org.xmloperator.lambda2.net.model.Cell;
import org.xmloperator.lambda2.net.model.Port;
import org.xmloperator.lambda2.net.model.Stack;

/**
 * Utilities about duplicators.
 */
public abstract class Duplicators {

	/**
	 * Extracts the duplicator of an interaction net.
	 * 
	 * @param support
	 *            the support of an interaction net.
	 * @return the multiplexed duplicator or null if the net doesn't contain a
	 *         duplicator.
	 */
	public static final BinaryCell extractDuplicator(Port support) {
		Stack duplicators = new StackImpl();
		collectDuplicators(support, duplicators);
		if (duplicators.isEmpty())
			return null;
		BinaryCellImpl duplicator0 = (BinaryCellImpl) duplicators.pop();
		if (duplicators.isEmpty())
			return duplicator0;
		Stack wireCellList = new StackImpl();
		addWireCells(duplicator0, wireCellList);
		PortImpl port0 = duplicator0;
		PortImpl port1 = duplicator0.port1;
		PortImpl port2 = duplicator0.port2;
		PortImpl connectedPort0 = port0.connectedPort;
		PortImpl connectedPort1 = port1.connectedPort;
		PortImpl connectedPort2 = port2.connectedPort;
		port0.disconnect();
		port1.disconnect();
		port2.disconnect();
		while (!duplicators.isEmpty()) {
			BinaryCellImpl duplicator = (BinaryCellImpl) duplicators.pop();
			addWireCells(duplicator, wireCellList);
			BinaryCellImpl constructor0 = new BinaryCellImpl(false);
			BinaryCellImpl constructor1 = new BinaryCellImpl(false);
			BinaryCellImpl constructor2 = new BinaryCellImpl(false);
			constructor0.connectTo(port0);
			constructor1.connectTo(port1);
			constructor2.connectTo(port2);
			port0 = constructor0.port1;
			port1 = constructor1.port2;
			port2 = constructor2.port2;
			constructor0.port2.connectTo(duplicator.connectedPort);
			constructor1.port1.connectTo(duplicator.port1.connectedPort);
			constructor2.port1.connectTo(duplicator.port2.connectedPort);
		}
		port0.connectTo(connectedPort0);
		port1.connectTo(connectedPort1);
		port2.connectTo(connectedPort2);
		while (!wireCellList.isEmpty()) {
			WireCellImpl wireCell = (WireCellImpl) wireCellList.pop();
			wireCell.connectedPort.connectTo(wireCell.port1.connectedPort);
		}
		return duplicator0;
	}

	/**
	 * Collects the duplicators of an interaction net.
	 * 
	 * @param support
	 *            the support of an interaction net.
	 * @param duplicators
	 *            receives the set of cells that implement the duplicators.
	 */
	private static final void collectDuplicators(Port support,
			Stack duplicators) {
		Set cells = new HashSet();
		Stack portStack = new StackImpl();
		portStack.push(support);
		while (!portStack.isEmpty()) {
			PortImpl port = (PortImpl) portStack.pop();
			PortImpl connectedPort = port.connectedPort;
			if (connectedPort == null)
				continue;
			Cell cell = connectedPort.cell;
			if (cell == null || !cells.add(cell))
				continue;
			if (cell.getSymbol() == Cell.SYMBOL_DUPLICATOR)
				duplicators.push(cell);
			cell.collectPorts(portStack);
		}
	}

	private static final void addWireCells(BinaryCellImpl duplicator,
			Stack wireCellList) {
		PortImpl port0 = duplicator;
		PortImpl port1 = duplicator.port1;
		PortImpl port2 = duplicator.port2;
		PortImpl connectedPort0 = port0.connectedPort;
		PortImpl connectedPort1 = port1.connectedPort;
		PortImpl connectedPort2 = port2.connectedPort;
		if (connectedPort0 != null) {
			Cell connectedCell = connectedPort0.cell;
			if (connectedCell != null
					&& connectedCell.getSymbol() == Cell.SYMBOL_DUPLICATOR) {
				WireCellImpl wireCell = new WireCellImpl();
				wireCellList.push(wireCell);
				wireCell.connectTo(connectedPort0);
				wireCell.port1.connectTo(port0);
			}
		}
		if (connectedPort1 != null) {
			Cell connectedCell = connectedPort1.cell;
			if (connectedCell != null
					&& connectedCell.getSymbol() == Cell.SYMBOL_DUPLICATOR) {
				WireCellImpl wireCell = new WireCellImpl();
				wireCellList.push(wireCell);
				wireCell.port1.connectTo(port1.connectedPort);
				wireCell.connectTo(port1);
			}
		}
		if (connectedPort2 != null) {
			Cell connectedCell = connectedPort2.getCell();
			if (connectedCell != null
					&& connectedCell.getSymbol() == Cell.SYMBOL_DUPLICATOR) {
				WireCellImpl wireCell = new WireCellImpl();
				wireCellList.push(wireCell);
				wireCell.port1.connectTo(port2.connectedPort);
				wireCell.connectTo(port2);
			}
		}
	}
}
