/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.net.impl;

import java.util.HashMap;
import java.util.Map;

import org.xmloperator.lambda2.net.model.IntStack;
import org.xmloperator.lambda2.net.model.Port;
import org.xmloperator.lambda2.net.model.Stack;
import org.xmloperator.lambda2.net.model.TooMuchReductionStepException;
import org.xmloperator.lambda2.term.model.Abstraction;
import org.xmloperator.lambda2.term.model.Application;
import org.xmloperator.lambda2.term.model.Term;
import org.xmloperator.lambda2.term.model.TermReducer;

/**
 * TermReducer implementation.
 */
public final class TermReducerImpl implements TermReducer {
	private static final int NODE_ABSTRACTION = 0;
	private static final int NODE_FUNCTION = 1;
	private static final int NODE_ARGUMENT = 2;

	private static final LambdaNetImpl lambdaNet(Term term) {
		Map fromVariableToSupport = new HashMap();
		IntStack buildingStack = new IntStackImpl();
		Stack termStack = new StackImpl();
		Stack argumentStack = new StackImpl();
		while (true) {
			LambdaNetImpl result = null;
			switch (term.getTermType()) {
				case Term.TYPE_VARIABLE : {
					LambdaNetImpl lambdaNet = new LambdaNetImpl();
					fromVariableToSupport.put(term,
							lambdaNet.bodySupport.connectedPort);
					result = lambdaNet;
					break;
				}
				case Term.TYPE_ABSTRACTION : {
					Abstraction abstraction = (Abstraction) term;
					term = abstraction.getBody();
					buildingStack.push(NODE_ABSTRACTION);
					termStack.push(abstraction);
					continue;
				}
				case Term.TYPE_APPLICATION : {
					Application application = (Application) term;
					if (application.getFunctionExponent() > Integer.MAX_VALUE)
						throw new IllegalArgumentException("Too big exponent");
					term = application.getArgument();
					buildingStack.push(NODE_ARGUMENT);
					termStack.push(application);
					continue;
				}
				default :
					throw new IllegalArgumentException();
			}
			while (true) {
				if (buildingStack.isEmpty()) {
					if (!fromVariableToSupport.isEmpty())
						throw new IllegalArgumentException("Free variables(s)");
					return result;
				}
				switch (buildingStack.pop()) {
					case NODE_ABSTRACTION : {
						Abstraction abstraction = (Abstraction) termStack.pop();
						Port[] freeVariableSupports = new Port[abstraction
								.getBoundVariableCount()];
						for (int k = 0; k < freeVariableSupports.length; k++) {
							Port freeVariableSupport = (Port) fromVariableToSupport
									.remove(abstraction.getBoundVariable(k));
							if (freeVariableSupport == null)
								throw new IllegalArgumentException(
										"Missing bound variable");
							freeVariableSupports[k] = freeVariableSupport;
						}
						result.buildAbstraction(freeVariableSupports);
						continue;
					}
					case NODE_ARGUMENT : {
						argumentStack.push(result);
						Application application = (Application) termStack.pop();
						term = application.getFunction();
						buildingStack.push(NODE_FUNCTION);
						termStack.push(application);
						break;
					}
					case NODE_FUNCTION : {
						Application application = (Application) termStack.pop();
						int exponent = (int) application.getFunctionExponent();
						LambdaNetImpl argument = (LambdaNetImpl) argumentStack
								.pop();
						result.buildApplication(argument, exponent);
						continue;
					}
					default :
						throw new IllegalStateException();
				}
				break;
			}
		}
	}

	private final int maxStepCount;
	private final int[] interactionStat;

	/**
	 * Constructor of TermReducer.
	 * 
	 * @param interactionStat
	 *            an array of interaction type count. May be null.
	 */
	public TermReducerImpl(int maxStepCount, int[] interactionStat) {
		this.maxStepCount = maxStepCount;
		this.interactionStat = interactionStat;
	}

	// ///////////////////////////
	// TermReducer implementation
	// ///////////////////////////

	public Term normalForm(Term term) throws TooMuchReductionStepException {
		LambdaNetImpl lambdaNet = lambdaNet(term);
		return lambdaNet
				.readBack(true, this.maxStepCount, this.interactionStat);
	}

	public boolean booleanValue(Term booleanTerm) {
		LambdaNetImpl lambdaNet = lambdaNet(booleanTerm);
		return lambdaNet.booleanValue(this.interactionStat);
	}

	public long numeralValue(Term numeralTerm) {
		LambdaNetImpl lambdaNet = lambdaNet(numeralTerm);
		return lambdaNet.numeralValue(this.interactionStat);
	}
}
