/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.net.model;

import org.xmloperator.lambda2.term.model.Term;

/**
 * <p>
 * A LambdaNet is an interaction-net that represents a lambda term which is
 * linear relativety to each of its free variables.
 * </p>
 * <p>
 * The net representation of the lambda term is the one of "Compiling the
 * lambda-calculus into Interaction Combinators" by Ian Mackie and Jorge Sousa
 * Pinto.
 * </p>
 * <p>
 * The interaction net body is accessible through (supported by) a free Port.
 * </p>
 */
public interface LambdaNet {

	/**
	 * Returns the free Port that supports the interaction-net body.
	 * 
	 * @return the free Port that supports the interaction-net body.
	 */
	public Port getBodySupport();

	/**
	 * Returns the count of free variables.
	 * 
	 * @return the count of free variables.
	 */
	public int getFreeVariableCount();

	/**
	 * Transforms this LambdaNet in such a way that its represented lambda term
	 * "t" is replaced by the abstraction "\x.t", where "x" doesn't occur within
	 * t.
	 */
	public void buildAbstraction();

	/**
	 * Transforms this LambdaNet in such a way that its represented lambda term
	 * "t" is replaced by the abstraction "\x.t", where "x" is bound to the
	 * given variable of "t".
	 * 
	 * @param freeVariableSupport
	 *            a support of free variable.
	 */
	public void buildAbstraction(Port freeVariableSupport);

	/**
	 * Transforms this LambdaNet in such a way that its represented lambda term
	 * "t" is replaced by the abstraction "\x.t", where "x" is bound to the
	 * given variables of "t".
	 * 
	 * @param freeVariableSupports
	 *            a set of 0 or more support of free variable.
	 */
	public void buildAbstraction(Port[] freeVariableSupports);

	/**
	 * <p>
	 * Transforms this LambdaNet in such a way that its represented lambda term
	 * "t" is replaced by the application "tu" where "u" is the lambda term
	 * represented by the given LambdaNet.
	 * </p>
	 * <p>
	 * The argument LambdaNet is consumed by the operation and is no more
	 * usable.
	 * </p>
	 * 
	 * @param argument
	 *            an argument LambdaNet.
	 * 
	 * @throws IllegalArgumentException
	 *             if the argument LambdaNet is this LambdaNet.
	 */
	public void buildApplication(LambdaNet argument);

	/**
	 * <p>
	 * Transforms this LambdaNet in such a way that its represented lambda term
	 * "t" is replaced by the application "(t^n)u" where "u" is the lambda term
	 * represented by the given LambdaNet and "n" an exponent.
	 * </p>
	 * <p>
	 * The argument LambdaNet is consumed by the operation and is no more
	 * usable.
	 * </p>
	 * 
	 * @param argument
	 *            an argument LambdaNet.
	 * @param functionExponent
	 *            a function exponent
	 * 
	 * @throws IllegalArgumentException
	 *             if the argument LambdaNet is this LambdaNet or if the
	 *             function exponent is lesser than one.
	 */
	public void buildApplication(LambdaNet argument, int functionExponent);

	/**
	 * Reduces this LambdaNet until the interface has a normal form.
	 * <p>
	 * Post-condition: the body port is a principal Port or a free Port.
	 * </p>
	 * 
	 * @param maxStepCount
	 *            a maximun step count. A negative value means no limit.
	 * @param interactionStat
	 *            an array of interaction type count. May be null.
	 * @return the count of reducing step.
	 * @throws TooMuchReductionStepException
	 *             if the given maximum count of step is not sufficient to
	 *             obtain an interface normal form.
	 */
	public int reduceUntilInterfaceNormalForm(int maxStepCount,
			int[] interactionStat) throws TooMuchReductionStepException;

	/**
	 * Read back the Term represented by this LambdaNet.
	 * <p>
	 * Pre-condition: this LambdaNet must have no free variable.
	 * </p>
	 * 
	 * @param inNormalForm
	 *            if true then returns a Term in normal form.
	 * @param maxStepCount
	 *            a maximum step count. A negative value means no limit.
	 * @param interactionStat
	 *            an array of interaction type count. May be null.
	 * 
	 * @return the Term read back from this LambdaNet.
	 * @throws IllegalStateException
	 *             if this LambdaNet has free variable(s).
	 * @throws TooMuchReductionStepException
	 *             if the given maximum count of step is not sufficient to
	 *             obtain a normal form or an interface normal form.
	 */
	public Term readBack(boolean inNormalForm, int maxStepCount,
			int[] interactionStat) throws TooMuchReductionStepException;

	/**
	 * Returns the boolean value of the Term (K or F) represented by this
	 * LambdaNet.
	 * 
	 * @param interactionStat
	 *            an array of interaction type count. May be null.
	 * @return the boolean value of the Term (K or F) represented by this
	 *         LambdaNet.
	 * @throws IllegalStateException
	 *             if this LambdaNet doesn't represent a boolean value or if it
	 *             has free variable(s).
	 */
	public boolean booleanValue(int[] interactionStat);

	/**
	 * Returns the value modulo 2^63 of the Church numeral Term represented by
	 * this LambdaNet. Returns -1 if this LambdaNet doesn't represent a Church
	 * numeral.
	 * 
	 * @param interactionStat
	 *            an array of interaction type count. May be null.
	 * @return the value modulo 2^63 of the Church numeral Term represented by
	 *         this LambdaNet. Returns -1 if this LambdaNet doesn't represent a
	 *         Church numeral.
	 * @throws IllegalStateException
	 *             if this LambdaNet doesn't represent a numeral value or if it
	 *             has free variable(s).
	 */
	public long numeralValue(int[] interactionStat);
}