/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.net.model;

/**
 * A Port belongs to a Cell or is free. A free Port is used as interface/support
 * of an interaction net.
 * <p>
 * A Port is either connected to another Port or disconnected.
 * </p>
 */
public interface Port {

	/**
	 * Returns the Cell to which this port belongs to or null if this Port is
	 * free.
	 * 
	 * @return the Cell to which this port belongs to or null if this Port is
	 *         free.
	 */
	public Cell getCell();

	/**
	 * Returns true if this Port is free, i.e. if it doesn't belong to a Cell.
	 * 
	 * @return true if this Port is free, i.e. if it doesn't belong to a Cell.
	 */
	public boolean isFree();

	/**
	 * Returns the number of this port. The principal Port of a Cell has the
	 * number 0. The other Ports of a Cell are numbered from 1. A free Port has
	 * <tt>-1</tt> as port number.
	 * 
	 * @return the number of this port.
	 */
	public int getPortNumber();

	/**
	 * Returns true if this Port is principal, false elsewhere.
	 * 
	 * @return true if this Port is principal, false elsewhere.
	 */
	public boolean isPrincipal();

	/**
	 * Returns the Port to which this one is connected or null if it is not
	 * connected.
	 * 
	 * @return the Port to which this one is connected or null if it is not
	 *         connected.
	 */
	public Port getConnectedPort();

	/**
	 * Returns true if this Port is connected, false elsewhere.
	 * 
	 * @return true if this Port is connected, false elsewhere.
	 */
	public boolean isConnected();

	/**
	 * Connects this Port to another Port. If this Port is already connected
	 * then it is first disconnected. If the other Port is already connected
	 * then it is first disconnected.
	 * 
	 * @param port
	 *            another Port.
	 * @throws IllegalArgumentException
	 *             if the given argument is itself.
	 * @throws NullPointerException
	 *             if the given argument is null.
	 */
	public void connectTo(Port port);

	/**
	 * Disconnects this Port.
	 * <p>
	 * Post-condition: connectedPort == null.
	 * </p>
	 */
	public void disconnect();
}
