/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.term.model;

import java.util.Set;

/**
 * A term of the lambda-calculus.
 */
public interface Term {
	public static final int TYPE_VARIABLE = 0;
	public static final int TYPE_ABSTRACTION = 1;
	public static final int TYPE_APPLICATION = 2;

	/**
	 * Returns the type of this term.
	 * 
	 * @return the type of this term. It is one of the constants TYPE_*
	 */
	public int getTermType();

	/**
	 * Returns <tt>true</tt> if this Term is a Variable.
	 * 
	 * @return <tt>true</tt> if this Term is a Variable.
	 */
	public boolean isVariable();

	/**
	 * Returns <tt>true</tt> if this Term is an Abstraction.
	 * 
	 * @return <tt>true</tt> if this Term is an Abstraction.
	 */
	public boolean isAbstraction();

	/**
	 * Returns <tt>true</tt> if this Term is an Application.
	 * 
	 * @return <tt>true</tt> if this Term is an Application.
	 */
	public boolean isApplication();

	/**
	 * Returns the set of free variables of this Term.
	 * 
	 * @return the set of free variables of this Term.
	 */
	public Set freeVariables();

	/**
	 * Returns a text representation of this Term. Exemple(Delta): "\x1x2.x1x2".
	 * 
	 * @return a text representation of this Term. Exemple(Delta): "\x1x2.x1x2".
	 */
	public String toText();

	/**
	 * Builds and returns an Abstraction whose body is this Term and which has
	 * no bound variable.
	 * 
	 * @return the Built Abstraction.
	 */
	public Abstraction buildAbstraction();

	/**
	 * Builds and returns an Abstraction whose body is this Term and with the
	 * given bound variable.
	 * 
	 * @param boundVariable
	 *            a Variable used once within this Term.
	 * @return the Built Abstraction.
	 */
	public Abstraction buildAbstraction(Variable boundVariable);

	/**
	 * Builds and returns an Abstraction whose body is this Term and with the
	 * given bound variables.
	 * 
	 * @param boundVariables
	 *            a set of variables used each once within this Term.
	 * @return the Built Abstraction.
	 */
	public Abstraction buildAbstraction(Variable[] boundVariables);

	/**
	 * Builds and returns an Application whose function is this Term and the
	 * argument is given.
	 * 
	 * @param argument
	 *            a Term.
	 * @return the build Application.
	 */
	public Application buildApplication(Term argument);

	/**
	 * Builds and returns an Application whose function is this Term and the
	 * argument is given. The function exponent may also be given.
	 * 
	 * @param argument
	 *            a Term.
	 * @param functionExponent
	 *            a function exponent.
	 * @return the build Application.
	 */
	public Application buildApplication(Term argument, long functionExponent);
}
