/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.term.util;

import java.io.PrintStream;

import org.xmloperator.lambda2.term.impl.VariableImpl;
import org.xmloperator.lambda2.term.model.Abstraction;
import org.xmloperator.lambda2.term.model.Application;
import org.xmloperator.lambda2.term.model.Term;
import org.xmloperator.lambda2.term.model.TermReducer;
import org.xmloperator.lambda2.term.model.Variable;

/**
 * Numerals writeln in base 2.
 */
public abstract class Base2Numerals {

	/**
	 * Term that represents the function that returns "K" if the argument is
	 * "0", "F" elsewhere.
	 */
	public static final Abstraction ZERO = newZero();

	/**
	 * Term that represents the function that returns half the value of the
	 * argument.
	 */
	public static final Abstraction HALF = newHalf();

	/**
	 * Term that represents the function that returns twice the value of the
	 * argument.
	 */
	public static final Abstraction TWICE0 = newTwice(false);

	/**
	 * Term that represents the function that returns twice the value of the
	 * argument plus one.
	 */
	public static final Abstraction TWICE1 = newTwice(true);

	/**
	 * Term that represents the function that returns the successor of the value
	 * of the argument.
	 */
	public static final Application SUCCESSOR = newSuccessor();

	/**
	 * Term that represents the function g(n, r) = 3 * n + r, with r = 0, 1 or
	 * 2.
	 */
	public static final Application THREE_N_PLUS_R = new3nPr();

	private static final int SELECT_P = 1;
	private static final int SELECT_Z = 2;
	private static final int SELECT_M = 3;

	/**
	 * Creates and returns a Term that represents an integer.
	 * <ul>
	 * <li>N(n) = \t.t P(n) Z(n) M(n)</li>
	 * <li>P(n) = parity of n (K if even, F if odd)</li>
	 * <li>Z(n) = using 1 bit (K if 0 or 1, F if 2 or more)</li>
	 * <li>M(n) = If n >= 2 then N(n/2) else I</li>
	 * <li>0 = \t.t K K I</li>
	 * <li>1 = \t.t F K I</li>
	 * <li>2 = \t.t K F 1</li>
	 * <li>3 = \t.t F F 1</li>
	 * <li>...</li>
	 * </ul>
	 * 
	 * @param n
	 *            the integer to represent.
	 * @return the base 2 numeral Term for the given integer.
	 * @throws IllegalArgumentException
	 *             if the integer n is negative.
	 */
	public static final Abstraction newInt(final int n) {
		if (n < 0)
			throw new IllegalArgumentException();
		int b0 = n % 2;
		int h = n / 2;
		Variable t = new VariableImpl();
		return t.buildApplication(CommonTerms.bool(b0 == 0)).buildApplication(
				CommonTerms.bool(n == 0)).buildApplication(
				n != 0 ? newInt(h) : CommonTerms.I).buildAbstraction(t);
	}

	/**
	 * Returns the long value modulo 2^63 of the given base 2 numeral term.
	 * 
	 * @param term
	 *            a base 2 numeral Term.
	 * @return the long value modulo 2^63 of the given base 2 numeral term.
	 * @throws IllegalStateException
	 *             if the given Term doesn't represent a base 2 numeral Term.
	 */
	public static final long longValue(TermReducer termReducer, Term term) {
		long bit0 = termReducer.booleanValue(CommonTerms.selectorForTriple(
				SELECT_P).buildApplication(term)) ? 0L : 1L;
		boolean usingOneBit = termReducer.booleanValue(CommonTerms
				.selectorForTriple(SELECT_Z).buildApplication(term));
		return usingOneBit
				? bit0
				: ((longValue(termReducer, CommonTerms.selectorForTriple(
						SELECT_M).buildApplication(term)) << 1) + bit0)
						& Long.MAX_VALUE;
	}

	/**
	 * Creates and returns a Term that represents the function that returns "K"
	 * if the argument is "0", "F" elsewhere.
	 * <p>
	 * Formula is : "\t.P t (Z t) F".
	 * </p>
	 * 
	 * @return the Zero function of base 2 numeral.
	 */
	private static final Abstraction newZero() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Application bz = CommonTerms.selectorForTriple(SELECT_Z)
				.buildApplication(x2); // Z x2
		return CommonTerms.selectorForTriple(SELECT_P).buildApplication(x1)
				.buildApplication(bz).buildApplication(CommonTerms.F)
				.buildAbstraction(new Variable[]{x1, x2});
	}

	/**
	 * Creates and returns a Term that represents the function that returns half
	 * the value of the argument.
	 * <p>
	 * Formula is : "\t.Z t 0 (M t)".
	 * </p>
	 * 
	 * @return the Half function of base 2 numeral.
	 */
	private static final Abstraction newHalf() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Application last = CommonTerms.selectorForTriple(SELECT_M)
				.buildApplication(x2); // M x2
		return CommonTerms.selectorForTriple(SELECT_Z).buildApplication(x1)
				.buildApplication(newInt(0)).buildApplication(last)
				.buildAbstraction(new Variable[]{x1, x2});
	}

	/**
	 * Creates and returns a Term that represents the function that returns
	 * twice the value of the argument.
	 * <p>
	 * Formula is : "\t.Z t 0(\r.r p F t)" with p = F if plusOne, K elsewhere.
	 * </p>
	 * 
	 * @param plusOne
	 *            if true then adds one to the function result. The function
	 *            result becomes 2n+1 in place of 2n.
	 * @return the Twice function of base 2 numeral.
	 */
	public static final Abstraction newTwice(boolean plusOne) {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable r = new VariableImpl();
		Abstraction twice = r.buildApplication(CommonTerms.bool(!plusOne))
				.buildApplication(CommonTerms.F).buildApplication(x2)
				.buildAbstraction(r); // \r.r p F x2
		return CommonTerms.selectorForTriple(SELECT_Z).buildApplication(x1)
				.buildApplication(newInt(0)).buildApplication(twice)
				.buildAbstraction(new Variable[]{x1, x2});
	}

	/**
	 * Creates and returns a Term that represents the function that returns the
	 * successor of the value of the argument.
	 * <p>
	 * Equation: Next N(n) =
	 * </p>
	 * <ul>
	 * <li>If P(n) then (\t.t F Z(n) M(n)) -- 2p+1 for n = 2p</li>
	 * <li>else if Z(n) then 2 -- 2 for n = 1</li>
	 * <li>else (\r.r K F (Next M(n))) -- 2(p+1) for n = 2p+1</li>
	 * </ul>
	 * <p>
	 * Lambda term is: "Y G"
	 * </p>
	 * with G = \s.\n.(P n)(\t.t F (Z n)(M n))(Z n 2 (\r.r K F (s (M n))))
	 * 
	 * @param plusOne
	 *            if true then adds one to the function result. The function
	 *            result becomes 2n+1 in place of 2n.
	 * @return the Twice function of base 2 numeral.
	 */
	private static final Application newSuccessor() {
		Variable xs = new VariableImpl();
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable x3 = new VariableImpl();
		Variable x4 = new VariableImpl();
		Variable x5 = new VariableImpl();
		Variable t1 = new VariableImpl();
		Variable t2 = new VariableImpl();
		Application bZ1 = CommonTerms.selectorForTriple(SELECT_Z)
				.buildApplication(x2); // Z x2
		Application bM1 = CommonTerms.selectorForTriple(SELECT_M)
				.buildApplication(x3); // M x3
		Application bM2 = CommonTerms.selectorForTriple(SELECT_M)
				.buildApplication(x5); // M x5
		Abstraction bT1 = t1.buildApplication(CommonTerms.F).buildApplication(
				bZ1).buildApplication(bM1).buildAbstraction(t1);
		// \t.t F (Z x2)(M x3)
		Abstraction bT2 = t2.buildApplication(CommonTerms.K).buildApplication(
				CommonTerms.F).buildApplication(xs.buildApplication(bM2))
				.buildAbstraction(t2);
		// \r.r K F (s (M x5))
		Application bZ2 = CommonTerms.selectorForTriple(SELECT_Z)
				.buildApplication(x4).buildApplication(newInt(2))
				.buildApplication(bT2); // Z x4 2 (\r.r K F (s (M x5)))
		Abstraction bg = CommonTerms.selectorForTriple(SELECT_P)
				.buildApplication(x1).buildApplication(bT1).buildApplication(
						bZ2).buildAbstraction(
						new Variable[]{x1, x2, x3, x4, x5}).buildAbstraction(
						new Variable[]{xs});
		// \s.\x1x2x3x4x5.(P x1)(\t.t F(Z x2)(M x3))(Z x4 2(\r.r K F(s(M x5))))
		return CommonTerms.Y.buildApplication(bg); // Y G
	}

	/**
	 * Creates and returns a Term that represents the function g(n, r) = 3 * n +
	 * r, with r = 0, 1 or 2.
	 * <p>
	 * It is based on this recursion:
	 * </p>
	 * <ul>
	 * <li>G N R =</li>
	 * <li>-- if (Z N)</li>
	 * <li>---- if (P N)</li>
	 * <li>------ case R0: 0</li>
	 * <li>------ case R1: 1</li>
	 * <li>------ case R2: 2</li>
	 * <li>---- else</li>
	 * <li>------ case R0: 3</li>
	 * <li>------ case R1: 4</li>
	 * <li>------ case R2: 5</li>
	 * <li>-- else</li>
	 * <li>---- if (P N)</li>
	 * <li>------ case R0: D0(G(M N)R0)</li>
	 * <li>------ case R1: D1(G(M N)R0)</li>
	 * <li>------ case R2: D0(G(M N)R1)</li>
	 * <li>---- else</li>
	 * <li>------ case R0: D1(G(M N)R1)</li>
	 * <li>------ case R1: D0(G(M N)R2)</li>
	 * <li>------ case R2: D1(G(M N)R2)</li>
	 * </ul>
	 * the lambda term is: "Y \g.\n.\r.rec(g n r)".
	 * 
	 * @return a Term that represents the function g(n, r) = 3 * n + r, with r =
	 *         0, 1 or 2.
	 */
	private static final Application new3nPr() {
		Variable g0 = new VariableImpl();
		Variable g1 = new VariableImpl();
		Variable g2 = new VariableImpl();
		Variable g3 = new VariableImpl();
		Variable g4 = new VariableImpl();
		Variable g5 = new VariableImpl();
		Variable nz = new VariableImpl();
		Variable np1 = new VariableImpl();
		Variable np2 = new VariableImpl();
		Variable n0 = new VariableImpl();
		Variable n1 = new VariableImpl();
		Variable n2 = new VariableImpl();
		Variable n3 = new VariableImpl();
		Variable n4 = new VariableImpl();
		Variable n5 = new VariableImpl();
		Variable r1 = new VariableImpl();
		Variable r2 = new VariableImpl();
		Variable r3 = new VariableImpl();
		Variable r4 = new VariableImpl();
		Term br1 = r1.buildApplication(newInt(0)).buildApplication(newInt(1))
				.buildApplication(newInt(2));
		Term br2 = r2.buildApplication(newInt(3)).buildApplication(newInt(4))
				.buildApplication(newInt(5));
		Term bp1 = CommonTerms.selectorForTriple(SELECT_P)
				.buildApplication(np1).buildApplication(br1).buildApplication(
						br2);
		Term bm0 = CommonTerms.selectorForTriple(SELECT_M).buildApplication(n0); // M N
		Term bg0 = g0.buildApplication(bm0).buildApplication(
				CommonTerms.FIRST_OF_THREE);
		Term bd0 = newTwice(false).buildApplication(bg0); // D0 (G (M N) R0)
		Term bm1 = CommonTerms.selectorForTriple(SELECT_M).buildApplication(n1); // M N
		Term bg1 = g1.buildApplication(bm1).buildApplication(
				CommonTerms.FIRST_OF_THREE);
		// G (M N) R0
		Term bd1 = newTwice(true).buildApplication(bg1); // D1 (G (M N) R0)
		Term bm2 = CommonTerms.selectorForTriple(SELECT_M).buildApplication(n2); // M N
		Term bg2 = g2.buildApplication(bm2).buildApplication(
				CommonTerms.SECOND_OF_THREE);
		// G (M N) R1
		Term bd2 = newTwice(false).buildApplication(bg2); // D0 (G (M N) R1)
		Term br3 = r3.buildApplication(bd0).buildApplication(bd1)
				.buildApplication(bd2);
		Term bm3 = CommonTerms.selectorForTriple(SELECT_M).buildApplication(n3); // M N
		Term bg3 = g3.buildApplication(bm3).buildApplication(
				CommonTerms.SECOND_OF_THREE);
		// G (M N) R1
		Term bd3 = newTwice(true).buildApplication(bg3); // D1 (G (M N) R1)
		Term bm4 = CommonTerms.selectorForTriple(SELECT_M).buildApplication(n4); // M N
		Term bg4 = g4.buildApplication(bm4).buildApplication(
				CommonTerms.THIRD_OF_THREE);
		// G (M N) R2
		Term bd4 = newTwice(false).buildApplication(bg4); // D0 (G (M N) R2)
		Term bm5 = CommonTerms.selectorForTriple(SELECT_M).buildApplication(n5); // M N
		Term bg5 = g5.buildApplication(bm5).buildApplication(
				CommonTerms.THIRD_OF_THREE);
		// G (M N) R2
		Term bd5 = newTwice(true).buildApplication(bg5); // D1 (G (M N) R2)
		Term br4 = r4.buildApplication(bd3).buildApplication(bd4)
				.buildApplication(bd5);
		Term bp2 = CommonTerms.selectorForTriple(SELECT_P)
				.buildApplication(np2).buildApplication(br3).buildApplication(
						br4);
		Term bz = CommonTerms.selectorForTriple(SELECT_Z).buildApplication(nz)
				.buildApplication(bp1).buildApplication(bp2).buildAbstraction(
						new Variable[]{r1, r2, r3, r4}).buildAbstraction(
						new Variable[]{nz, np1, np2, n0, n1, n2, n3, n4, n5})
				.buildAbstraction(new Variable[]{g0, g1, g2, g3, g4, g5});
		return CommonTerms.Y.buildApplication(bz);
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final boolean IS_VERBOSE = false;
	private static final TermReducer TERM_REDUCER = Test.TERM_REDUCER;

	/**
	 * Execute some operations using base 2 numerals.
	 * 
	 * @param out
	 *            the PrintStream to print results. May be null.
	 * @param isVerbose
	 *            if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			boolean isOk = true;
			isOk &= test(out, isVerbose, newInt(0), 0, "0");
			isOk &= test(out, isVerbose, newInt(1), 1, "1");
			isOk &= test(out, isVerbose, newInt(2), 2, "2");
			isOk &= test(out, isVerbose, newInt(3), 3, "3");
			isOk &= test(out, isVerbose, CommonTerms
					.selectorForTriple(SELECT_P).buildApplication(newInt(0)),
					true, "Parity 0");
			isOk &= test(out, isVerbose, CommonTerms
					.selectorForTriple(SELECT_P).buildApplication(newInt(1)),
					false, "Parity 1");
			isOk &= test(out, isVerbose, CommonTerms
					.selectorForTriple(SELECT_P).buildApplication(newInt(2)),
					true, "Parity 2");
			isOk &= test(out, isVerbose, CommonTerms
					.selectorForTriple(SELECT_P).buildApplication(newInt(3)),
					false, "Parity 3");
			isOk &= test(out, isVerbose, ZERO.buildApplication(newInt(0)),
					true, "Zero 0");
			isOk &= test(out, isVerbose, ZERO.buildApplication(newInt(1)),
					false, "Zero 1");
			isOk &= test(out, isVerbose, ZERO.buildApplication(newInt(2)),
					false, "Zero 2");
			isOk &= test(out, isVerbose, HALF.buildApplication(newInt(0)), 0,
					"Half 0");
			isOk &= test(out, isVerbose, HALF.buildApplication(newInt(1)), 0,
					"Half 1");
			isOk &= test(out, isVerbose, HALF.buildApplication(newInt(2)), 1,
					"Half 2");
			isOk &= test(out, isVerbose, HALF.buildApplication(newInt(3)), 1,
					"Half 3");
			isOk &= test(out, isVerbose, TWICE0.buildApplication(newInt(3)), 6,
					"Twice 3");
			isOk &= test(out, isVerbose, TWICE1.buildApplication(newInt(3)), 7,
					"Twice+1 3");
			isOk &= test(out, isVerbose, SUCCESSOR.buildApplication(newInt(0)),
					1, "S 0");
			isOk &= test(out, isVerbose, SUCCESSOR.buildApplication(newInt(1)),
					2, "S 1");
			isOk &= test(out, isVerbose, SUCCESSOR.buildApplication(newInt(2)),
					3, "S 2");
			isOk &= test(out, isVerbose, SUCCESSOR.buildApplication(newInt(3)),
					4, "S 3");
			isOk &= test(out, isVerbose, THREE_N_PLUS_R.buildApplication(
					newInt(3)).buildApplication(CommonTerms.THIRD_OF_THREE),
					11, "3 * 3 + 2");
			return isOk;
		} catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static boolean test(PrintStream out, boolean isVerbose, Term term,
			boolean expectedValue, String name) {
		boolean value = TERM_REDUCER.booleanValue(term);
		boolean isOk = value == expectedValue;
		if (isVerbose || !isOk)
			out.println("Base2Numerals " + (isOk ? "ok" : "ko") + " " + value
					+ " / " + expectedValue + " (" + name + ")" + " net: ");
		return isOk;
	}

	private static boolean test(PrintStream out, boolean isVerbose, Term term,
			long expectedValue, String name) {
		long value = longValue(TERM_REDUCER, term);
		boolean isOk = value == expectedValue;
		if (isVerbose || !isOk)
			out.println("Base2Numerals " + (isOk ? "ok" : "ko") + " " + value
					+ " / " + expectedValue + " (" + name + ")" + " net: ");
		return isOk;
	}
}
