/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.term.util;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import org.xmloperator.lambda2.term.impl.VariableImpl;
import org.xmloperator.lambda2.term.model.Abstraction;
import org.xmloperator.lambda2.term.model.Application;
import org.xmloperator.lambda2.term.model.Term;
import org.xmloperator.lambda2.term.model.TermReducer;
import org.xmloperator.lambda2.term.model.Variable;

/**
 * Church numerals.
 */
public abstract class ChurchNumerals {
	private static final boolean IS_FUNCTION_EXPONENT_ENABLED = false;
	
	/**
	 * Term <tt>\x1x2x3.x2(x1x2x3)</tt>
	 */
	public static final Abstraction SUCC = newSucc();

	/**
	 * Term <tt>\x1x2x3.x1(\x4.x3)x2</tt>
	 */
	public static final Abstraction ZERO = newZero();

	/**
	 * Term <tt>\x1x2x3x4.x1x3(x2x3x4)</tt>
	 */
	public static final Abstraction ADD = newAdd();

	/**
	 * Term <tt>\x1x2x3.x1(x2x3)</tt>
	 */
	public static final Abstraction MULT = newMult();

	/**
	 * Term <tt>\x1x2.x2x1</tt>
	 */
	public static final Abstraction EXP = newExp();

	/**
	 * Creates and returns a Term that represents an integer.
	 * <ul>
	 * <li>0 = "\x1.\x2.x2",</li>
	 * <li>1 = "\x1.\x2.x1x2",</li>
	 * <li>2 = "\x1.\x2.x1(x1x2)",</li>
	 * <li>3 = "\x1.\x2.x1(x1(x1x2))",</li>
	 * <li>...</li>
	 * </ul>
	 * 
	 * @param n
	 *            the integer to represent.
	 * @return the Church numeral Term for the given integer.
	 * @throws IllegalArgumentException
	 *             if the integer n is negative.
	 */
	public static final Abstraction newInt(int n) {
		if (n < 0)
			throw new IllegalArgumentException();
		Variable x2 = new VariableImpl();
		if (n == 0)
			return x2.buildAbstraction(x2).buildAbstraction();
		if (IS_FUNCTION_EXPONENT_ENABLED) {
			Variable x1 = new VariableImpl();
			return x1.buildApplication(x2, n).buildAbstraction(x2)
					.buildAbstraction(x1);
		} else {
			List x1List = new ArrayList();
			Term x = x2;
			for (int k = 0; k < n; k++) {
				Variable x1 = new VariableImpl();
				x1List.add(x1);
				x = x1.buildApplication(x);
			}
			Variable[] x1Supports = new Variable[x1List.size()];
			x1List.toArray(x1Supports);
			return x.buildAbstraction(x2).buildAbstraction(x1Supports);
		}
	}

	/**
	 * Creates and returns a "Succ" Term <tt>"\x1.\x2.\x3.x2(x1x2x3)"</tt>.
	 * 
	 * @return the "Succ" Term.
	 */
	private static final Abstraction newSucc() {
		Variable x1 = new VariableImpl();
		Variable x21 = new VariableImpl();
		Variable x22 = new VariableImpl();
		Variable x3 = new VariableImpl();
		Application p = x1.buildApplication(x22).buildApplication(x3);
		return x21.buildApplication(p).buildAbstraction(x3).buildAbstraction(
				new Variable[]{x21, x22}).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a "Zero" Term <tt>"\x1.\x2.\x3.x1(\x4.x3)x2"</tt>.
	 * 
	 * @return a "Zero" Term.
	 */
	private static final Abstraction newZero() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable x3 = new VariableImpl();
		Abstraction b3 = x3.buildAbstraction(); // \x4.x3
		return x1.buildApplication(b3).buildApplication(x2)
				.buildAbstraction(x3).buildAbstraction(x2).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a "Add" Term <tt>"\x1.\x2.\x3.\x4.x1x3(x2x3x4)"</tt>.
	 * 
	 * @return a "Add" Term.
	 */
	private static final Abstraction newAdd() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable x31 = new VariableImpl();
		Variable x32 = new VariableImpl();
		Variable x4 = new VariableImpl();
		Application p = x2.buildApplication(x32).buildApplication(x4);
		return x1.buildApplication(x31).buildApplication(p)
				.buildAbstraction(x4)
				.buildAbstraction(new Variable[]{x31, x32})
				.buildAbstraction(x2).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a "Mult" Term <tt>"\x1.\x2.\x3.x1(x2x3)"</tt>.
	 * 
	 * @return a "Mult" Term.
	 */
	private static final Abstraction newMult() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable x3 = new VariableImpl();
		Application p = x2.buildApplication(x3); // x2x3
		return x1.buildApplication(p).buildAbstraction(x3).buildAbstraction(x2)
				.buildAbstraction(x1);
	}

	/**
	 * Creates and returns a "Exp" Term <tt>"\x1.\x2.x2x1"</tt>.
	 * 
	 * @return a "Exp" Term.
	 */
	private static final Abstraction newExp() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		return x2.buildApplication(x1).buildAbstraction(x2)
				.buildAbstraction(x1);
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final boolean IS_VERBOSE = false;
	private static final TermReducer TERM_REDUCER = Test.TERM_REDUCER;

	/**
	 * Execute some operations using Church numerals.
	 * 
	 * @param out
	 *            the PrintStream to print results. May be null.
	 * @param isVerbose
	 *            if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			boolean isOk = true;
			isOk &= test(out, isVerbose, newInt(0), "\\x1.\\x2.x2", "0");
			isOk &= test(out, isVerbose, newInt(1), "\\x1.\\x2.x1x2", "1");
			isOk &= test(out, isVerbose, newInt(2), "\\x1.\\x2.x1(x1x2)", "2");
			isOk &= test(out, isVerbose, EXP, "\\x1.\\x2.x2x1", "Exp");
			isOk &= test(out, isVerbose, MULT, "\\x1.\\x2.\\x3.x1(x2x3)",
					"Mult");
			isOk &= test(out, isVerbose, ADD,
					"\\x1.\\x2.\\x3.\\x4.x1x3(x2x3x4)", "Add");
			isOk &= test(out, isVerbose, ZERO, "\\x1.\\x2.\\x3.(x1\\x4.x3)x2",
					"Zero");
			isOk &= test(out, isVerbose, SUCC, "\\x1.\\x2.\\x3.x2(x1x2x3)",
					"Succ");
			isOk &= test(out, isVerbose, newInt(0), 0, "0");
			isOk &= test(out, isVerbose, newInt(1), 1, "1");
			isOk &= test(out, isVerbose, newInt(2), 2, "2");
			isOk &= test(out, isVerbose, newInt(3), 3, "3");
			isOk &= test(out, isVerbose, newInt(4), 4, "4");
			isOk &= test(out, isVerbose, ZERO.buildApplication(newInt(0)),
					true, "Zero 0");
			isOk &= test(out, isVerbose, ZERO.buildApplication(newInt(1)),
					false, "Zero 1");
			isOk &= test(out, isVerbose, ZERO.buildApplication(newInt(2)),
					false, "Zero 2");
			isOk &= test(out, isVerbose, newInt(2).buildApplication(newInt(2)),
					4, "2 2");
			isOk &= test(out, isVerbose, SUCC.buildApplication(newInt(0)), 1,
					"Succ 0");
			isOk &= test(out, isVerbose, SUCC.buildApplication(newInt(1)), 2,
					"Succ 1");
			isOk &= test(out, isVerbose, SUCC.buildApplication(newInt(2)), 3,
					"Succ 2");
			isOk &= test(out, isVerbose, ADD.buildApplication(newInt(3))
					.buildApplication(newInt(3)), 6, "Add 3 3");
			isOk &= test(out, isVerbose, MULT.buildApplication(newInt(3))
					.buildApplication(newInt(2)), 6, "Mult 3 2");
			isOk &= test(out, isVerbose, EXP.buildApplication(newInt(2))
					.buildApplication(newInt(2)), 4, "Exp 2 2");
			isOk &= test(out, isVerbose, newInt(3).buildApplication(newInt(2)),
					8, "3 2");
			isOk &= test(out, isVerbose, newInt(8).buildApplication(newInt(2)),
					256, "8 2");
			{
				int n = 60;
				long t0 = System.currentTimeMillis();
				Term term = newInt(n).buildApplication(newInt(2));
				if (isVerbose)
					out.println("ChurchNumerals 2^" + n);
				long result = TERM_REDUCER.numeralValue(term);
				long t1 = System.currentTimeMillis();
				long expectedResult = n < 64 ? ((1L << n) & Long.MAX_VALUE) : 0;
				boolean isResultOk = result == expectedResult;
				if (isVerbose || !isResultOk) {
					out.println("ChurchNumerals " + (isResultOk ? "ok" : "ko")
							+ " 2^" + n + " = " + result + " / "
							+ expectedResult);
					isOk &= isResultOk;
				}
				if (isVerbose && isResultOk)
					out.println("ChurchNumerals intValue(2^" + n + ") using "
							+ (t1 - t0) + " ms");
			}
			{
				int n = 8;
				long t0 = System.currentTimeMillis();
				Term term = newInt(n).buildApplication(newInt(2))
						.buildApplication(newInt(2)).buildApplication(
								CommonTerms.I).buildApplication(CommonTerms.I);
				if (isVerbose)
					out.println("ChurchNumerals 2^(2^" + n + ") I I");
				TERM_REDUCER.numeralValue(term);
				long t1 = System.currentTimeMillis();
				if (isVerbose)
					out.println("ChurchNumerals intValue(2^(2^" + n
							+ ") I I) using " + (t1 - t0) + " ms");
			}
			return isOk;
		} catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static boolean test(PrintStream out, boolean isVerbose, Term term,
			String expectedText, String name) {
		String text = term.toText();
		boolean isOk = text.equals(expectedText);
		if (isVerbose || !isOk)
			out.println("ChurchNumerals " + (isOk ? "ok" : "ko") + " \"" + text
					+ "\" / \"" + expectedText + "\" (" + name + ")");
		return isOk;
	}

	private static boolean test(PrintStream out, boolean isVerbose, Term term,
			boolean expectedValue, String name) {
		String text = TERM_REDUCER.normalForm(term).toText();
		boolean value = TERM_REDUCER.booleanValue(term);
		boolean isOk = value == expectedValue;
		if (isVerbose || !isOk) {
			out.println("ChurchNumerals " + (isOk ? "ok" : "ko") + " " + value
					+ " / " + expectedValue + " (" + name + ")" + " term: "
					+ text);
		}
		return isOk;
	}

	private static boolean test(PrintStream out, boolean isVerbose, Term term,
			long expectedValue, String name) {
		String text = TERM_REDUCER.normalForm(term).toText();
		long value = TERM_REDUCER.numeralValue(term);
		boolean isOk = value == expectedValue;
		if (isVerbose || !isOk) {
			out.println("ChurchNumerals " + (isOk ? "ok" : "ko") + " " + value
					+ " / " + expectedValue + " (" + name + ")" + " term: "
					+ text);
		}
		return isOk;
	}
}