/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.term.util;

import java.io.PrintStream;

import org.xmloperator.lambda2.term.impl.VariableImpl;
import org.xmloperator.lambda2.term.model.Abstraction;
import org.xmloperator.lambda2.term.model.Application;
import org.xmloperator.lambda2.term.model.Term;
import org.xmloperator.lambda2.term.model.TermReducer;
import org.xmloperator.lambda2.term.model.Variable;

/**
 * Common Terms.
 */
public abstract class CommonTerms {

	/**
	 * Term <tt>\x1.x1</tt>
	 */
	public static final Abstraction I = newIdentity();

	/**
	 * Term <tt>\x1x2.x1</tt>
	 */
	public static final Abstraction K = newBool(true);

	/**
	 * Term <tt>\x1x2.x2</tt>
	 */
	public static final Abstraction F = newBool(false);

	/**
	 * Term <tt>\x1x2x3.x1</tt>
	 */
	public static final Abstraction FIRST_OF_THREE = newThreeStates(1);

	/**
	 * Term <tt>\x1x2x3.x2</tt>
	 */
	public static final Abstraction SECOND_OF_THREE = newThreeStates(2);

	/**
	 * Term <tt>\x1x2x3.x3</tt>
	 */
	public static final Abstraction THIRD_OF_THREE = newThreeStates(3);

	/**
	 * Term <tt>\x1x2x3.x1x2x3</tt>
	 */
	public static final Abstraction IF = newIf();

	/**
	 * Term <tt>\x1.x1FK</tt>
	 */
	public static final Abstraction NOT = newNot();

	/**
	 * Term <tt>\x1x2.x1x2F</tt>
	 */
	public static final Abstraction AND = newAnd();

	/**
	 * Term <tt>\x1x2.x1Kx2</tt>
	 */
	public static final Abstraction OR = newOr();

	/**
	 * Term <tt>\x1x2x3.x3x1x2</tt>
	 */
	public static final Abstraction COUPLE = newCouple();

	/**
	 * Term <tt>\x1.x1\x2x3.x2</tt>
	 */
	public static final Abstraction FIRST_IN_COUPLE = newSelectorForCouple(true);

	/**
	 * Term <tt>\x1.x1\x2x3.x3</tt>
	 */
	public static final Abstraction SECOND_IN_COUPLE = newSelectorForCouple(false);

	/**
	 * Term <tt>\x1x2x3x4.x4x1x2x3</tt>
	 */
	public static final Abstraction TRIPLE = newTriple();

	/**
	 * Term <tt>\x1.x1\x2x3x4.x2</tt>
	 */
	public static final Abstraction FIRST_IN_TRIPLE = newSelectorForTriple(1);

	/**
	 * Term <tt>\x1.x1\x2x3x4.x3</tt>
	 */
	public static final Abstraction SECOND_IN_TRIPLE = newSelectorForTriple(2);

	/**
	 * Term <tt>\x1.x1\x2x3x4.x4</tt>
	 */
	public static final Abstraction THIRD_IN_TRIPLE = newSelectorForTriple(3);

	/**
	 * Term Y.
	 * <p>
	 * <tt>\x1.(\x2.x1(x2x2))\x2.x1(x2x2)</tt>
	 * </p>
	 */
	public static final Abstraction Y = newY();

	/**
	 * Term <tt>\x1x2.x2(x1x1x2)</tt>
	 */
	public static final Abstraction V = newV();

	/**
	 * Term <tt>(\x1x2.x2(x1x1x2))\x1x2.x2(x1x1x2)</tt>
	 */
	public static final Application TETA = V.buildApplication(V);

	/**
	 * Term <tt>\x1.x1x1</tt>
	 */
	public static final Abstraction DELTA = newDelta();

	/**
	 * Term <tt>(\x1.x1x1)\x1.x1x1</tt>
	 */
	public static final Application OMEGA = DELTA.buildApplication(DELTA);

	/**
	 * Creates and returns a Term that represents a boolean value.
	 * 
	 * @param value
	 *            a boolean value.
	 * @return a boolean selector Term.
	 */
	public static final Abstraction bool(boolean value) {
		return value ? K : F;
	}

	/**
	 * Creates and returns a Term that represents a selector amongst three
	 * states.
	 * 
	 * @param value
	 *            1, 2 or 3 (for first, second and last).
	 * @return a three states selector Term.
	 */
	public static final Abstraction threeStates(int value) {
		switch (value) {
			case 1 :
				return FIRST_OF_THREE;
			case 2 :
				return SECOND_OF_THREE;
			case 3 :
				return THIRD_OF_THREE;
			default :
				throw new IllegalArgumentException();
		}
	}

	/**
	 * Creates and returns a Term that represents a selector of a term in a
	 * couple : First = "\x1.x1\x2x3.x2" : Last = "\x1.x1\x2x3.x3".
	 * 
	 * @param selector :
	 *            true for First, false for Last.
	 * @return a Term <tt>First</tt> or <tt>Last</tt> in a couple.
	 */
	public static final Abstraction selectorForCouple(boolean selector) {
		return selector ? FIRST_IN_COUPLE : SECOND_IN_COUPLE;
	}

	/**
	 * Creates and returns a Term that represents a selector of a term in a
	 * triple :
	 * <ul>
	 * <li>First = "\x1.x1\x2x3x4.x2"</li>
	 * <li>Second = "\x1.x1\x2x3x4.x3"</li>
	 * <li>Last = "\x1.x1\x2x3x4.x4"</li>
	 * </ul>
	 * 
	 * @param selector :
	 *            1 for First, 2 for Second, 3 for Last.
	 * @return a selector of the first term in a triple.
	 */
	public static final Abstraction selectorForTriple(int selector) {
		switch (selector) {
			case 1 :
				return FIRST_IN_TRIPLE;
			case 2 :
				return SECOND_IN_TRIPLE;
			case 3 :
				return THIRD_IN_TRIPLE;
			default :
				throw new IllegalArgumentException();
		}
	}

	/**
	 * Creates and returns a Term <tt>"I" = "\x1.x1"</tt>.
	 * 
	 * @return the identity term.
	 */
	private static final Abstraction newIdentity() {
		VariableImpl x1 = new VariableImpl();
		return x1.buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term that represents a boolean selector.
	 * 
	 * @param value
	 *            a boolean value.
	 * @return a boolean selector Term.
	 */
	private static final Abstraction newBool(boolean value) {
		Variable xk = new VariableImpl();
		return value ? xk.buildAbstraction().buildAbstraction(xk) : xk
				.buildAbstraction(xk).buildAbstraction();
	}

	/**
	 * Creates and returns a Term that represents a selector amongst three
	 * states.
	 * 
	 * @param value
	 *            1, 2 or 3 (for first, second and last).
	 * @return a three states selector Term.
	 */
	private static final Abstraction newThreeStates(int value) {
		Variable xk = new VariableImpl();
		switch (value) {
			case 1 :
				return xk.buildAbstraction().buildAbstraction()
						.buildAbstraction(xk);
			case 2 :
				return xk.buildAbstraction().buildAbstraction(xk)
						.buildAbstraction();
			case 3 :
				return xk.buildAbstraction(xk).buildAbstraction()
						.buildAbstraction();
			default :
				throw new IllegalArgumentException();
		}
	}

	/**
	 * Creates and returns a Term <tt>"If" = "\x1.\x2.\x3.x1x2x3"</tt>.
	 * 
	 * @return an <tt>If</tt> Term.
	 */
	private static final Abstraction newIf() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable x3 = new VariableImpl();
		return x1.buildApplication(x2).buildApplication(x3)
				.buildAbstraction(x3).buildAbstraction(x2).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term <tt>"Not" = "\x1.x1FK"</tt>.
	 * 
	 * @return a <tt>Not</tt> Term.
	 */
	private static final Abstraction newNot() {
		Variable x1 = new VariableImpl();
		return x1.buildApplication(F).buildApplication(K).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term <tt>"And" = "\x1.\x2.x1x2F"</tt>.
	 * 
	 * @return a <tt>And</tt> Term.
	 */
	private static final Abstraction newAnd() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		return x1.buildApplication(x2).buildApplication(F).buildAbstraction(x2)
				.buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term <tt>"Or" = "\x1.\x2.x1Kx2"</tt>.
	 * 
	 * @return a <tt>Or</tt> Term.
	 */
	private static final Abstraction newOr() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		return x1.buildApplication(K).buildApplication(x2).buildAbstraction(x2)
				.buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term <tt>"C" = "\x1.\x2.\x3.x3x1x2"</tt>.
	 * 
	 * @return a <tt>C</tt> Term.
	 */
	private static final Abstraction newCouple() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable x3 = new VariableImpl();
		return x3.buildApplication(x1).buildApplication(x2)
				.buildAbstraction(x3).buildAbstraction(x2).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term that represents a selector of a term in a
	 * couple : First = "\x1.x1\x2.\x3.x2" : Last = "\x1.x1\x2.\x3.x3".
	 * 
	 * @param selector :
	 *            true for First, false for Last.
	 * @return a Term <tt>First</tt> or <tt>Last</tt> in a couple.
	 */
	private static final Abstraction newSelectorForCouple(boolean selector) {
		Variable x1 = new VariableImpl();
		return x1.buildApplication(bool(selector)).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term <tt>"T" = "\x1.\x2.\x3.\x4.x4x1x2x3"</tt>.
	 * 
	 * @return a Term <tt>T</tt>.
	 */
	private static final Abstraction newTriple() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable x3 = new VariableImpl();
		Variable x4 = new VariableImpl();
		return x4.buildApplication(x1).buildApplication(x2)
				.buildApplication(x3).buildAbstraction(x4).buildAbstraction(x3)
				.buildAbstraction(x2).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term that represents a selector of a term in a
	 * triple :
	 * <ul>
	 * <li>First = "\x1.x1\x2.\x3.\x4.x2"</li>
	 * <li>Second = "\x1.x1\x2.\x3.\x4.x3"</li>
	 * <li>Last = "\x1.x1\x2.\x3.\x4.x4"</li>
	 * </ul>
	 * 
	 * @param selector :
	 *            1 for First, 2 for Second, 3 for Last.
	 * @return a selector of the first term in a triple.
	 */
	private static final Abstraction newSelectorForTriple(int selector) {
		Variable x1 = new VariableImpl();
		return x1.buildApplication(threeStates(selector)).buildAbstraction(x1);
	}

	/**
	 * Creates and returns a Term <tt>"\x2.x1(x2x2)"</tt>.
	 * 
	 * @return a <tt>Y</tt> part Term.
	 */
	private static final Abstraction newYPart(Variable x1) {
		Variable x21 = new VariableImpl();
		Variable x22 = new VariableImpl();
		return x1.buildApplication(x21.buildApplication(x22)).buildAbstraction(
				new Variable[]{x21, x22});
	}

	/**
	 * Creates and returns a Term
	 * <tt>"Y" = "\x1.(\x2.x1(x2x2))\x2.x1(x2x2)"</tt>.
	 * 
	 * @return a Term <tt>Y</tt>.
	 */
	private static final Abstraction newY() {
		Variable x11 = new VariableImpl();
		Variable x12 = new VariableImpl();
		return newYPart(x11).buildApplication(newYPart(x12)).buildAbstraction(
				new Variable[]{x11, x12});
	}

	/**
	 * Creates and returns a Term <tt>"\x1.\x2.x2(x1x1x2)"</tt>
	 * 
	 * @return a Term <tt>V</tt>.
	 */
	private static final Abstraction newV() {
		Variable x11 = new VariableImpl();
		Variable x12 = new VariableImpl();
		Variable x21 = new VariableImpl();
		Variable x22 = new VariableImpl();
		return x21.buildApplication(
				x11.buildApplication(x12).buildApplication(x22))
				.buildAbstraction(new Variable[]{x21, x22}).buildAbstraction(
						new Variable[]{x11, x12});
	}

	/**
	 * Creates and returns a Term <tt>"Delta" = "\x1.x1x1"</tt>.
	 * 
	 * @return a Term <tt>Delta</tt>.
	 */
	private static final Abstraction newDelta() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		return x1.buildApplication(x2).buildAbstraction(new Variable[]{x1, x2});
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final boolean IS_VERBOSE = false;
	private static final TermReducer TERM_REDUCER = Test.TERM_REDUCER;

	/**
	 * Try some common Builders.
	 * 
	 * @param out
	 *            the PrintStream to print results. May be null.
	 * @param isVerbose
	 *            if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			boolean isOk = true;
			isOk &= test(out, isVerbose, I, "\\x1.x1", "I");
			isOk &= test(out, isVerbose, K, "\\x1.\\x2.x1", "K");
			isOk &= test(out, isVerbose, F, "\\x1.\\x2.x2", "F");
			isOk &= test(out, isVerbose, IF, "\\x1.\\x2.\\x3.x1x2x3", "If");
			isOk &= test(out, isVerbose, NOT,
					"\\x1.(x1\\x2.\\x3.x3)\\x2.\\x3.x2", "Not");
			isOk &= test(out, isVerbose, TERM_REDUCER.normalForm(NOT
					.buildApplication(K)), "\\x1.\\x2.x2", "Not K");
			isOk &= test(out, isVerbose, TERM_REDUCER.normalForm(NOT
					.buildApplication(F)), "\\x1.\\x2.x1", "Not F");
			isOk &= test(out, isVerbose, AND, "\\x1.\\x2.x1x2\\x3.\\x4.x4",
					"And");
			isOk &= test(out, isVerbose, OR, "\\x1.\\x2.(x1\\x3.\\x4.x3)x2",
					"Or");
			isOk &= test(out, isVerbose, COUPLE, "\\x1.\\x2.\\x3.x3x1x2", "C");
			isOk &= test(out, isVerbose, FIRST_IN_COUPLE,
					"\\x1.x1\\x2.\\x3.x2", "Fst");
			isOk &= test(out, isVerbose, SECOND_IN_COUPLE,
					"\\x1.x1\\x2.\\x3.x3", "Snd");
			isOk &= test(out, isVerbose, Y,
					"\\x1.(\\x2.x1(x2x2))\\x2.x1(x2x2)", "Y");
			isOk &= test(out, isVerbose, TETA,
					"(\\x1.\\x2.x2(x1x1x2))\\x1.\\x2.x2(x1x1x2)", "Teta");
			isOk &= test(out, isVerbose, OMEGA, "(\\x1.x1x1)\\x1.x1x1", "Omega");
			return isOk;
		} catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static boolean test(PrintStream out, boolean isVerbose, Term term,
			String expectedText, String name) {
		String text = term.toText();
		boolean isOk = text.equals(expectedText);
		if (isVerbose || !isOk)
			out.println("CommonTerms " + (isOk ? "ok" : "ko") + " \"" + text
					+ "\" / \"" + expectedText + "\" (" + name + ")");
		return isOk;
	}
}