/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.term.util;

import java.io.PrintStream;

import org.xmloperator.lambda2.term.impl.VariableImpl;
import org.xmloperator.lambda2.term.model.Abstraction;
import org.xmloperator.lambda2.term.model.Application;
import org.xmloperator.lambda2.term.model.Term;
import org.xmloperator.lambda2.term.model.TermReducer;
import org.xmloperator.lambda2.term.model.Variable;

/**
 * Terms that represent couples (n / 2, n % 2).
 */
public abstract class Modulo2Numerals {

	/**
	 * Term that represents a function of a Church numeral that returns "K" if
	 * the numeral is even, "F" elsewhere.
	 */
	public static final Abstraction PARITY = newParity();

	/**
	 * Term that represents the successor function for a numeral under the form
	 * (n / 2, n % 2).
	 */
	public static final Abstraction MOD_SUCC = newModSucc();

	/**
	 * Term that represents a function of a Church numeral that returns half the
	 * value of the argument.
	 */
	public static final Abstraction HALF = newHalf();

	/**
	 * Creates and returns a Term that represents a numeral under the form (n /
	 * 2, n % 2) : \t.t churchInt(n / 2) boolean(n % 2).
	 * 
	 * @param n
	 *            the integer to represent.
	 * @return a Term for a numeral under the form (n / 2, n % 2)
	 */
	public static final Abstraction newModInt(int n) {
		int half = n / 2;
		boolean parity = n % 2 == 0;
		Variable t = new VariableImpl();
		return t.buildApplication(ChurchNumerals.newInt(half))
				.buildApplication(CommonTerms.bool(parity)).buildAbstraction(t);
	}

	/**
	 * Creates and returns a Term that represents a function of a Church numeral
	 * that returns "K" if the numeral is even, "F" elsewhere.
	 * <p>
	 * Formula: Parity n = n not K
	 * </p>
	 * 
	 * @return a Term that represents a function of a Church numeral that
	 *         returns "K" if the numeral is even, "F" if it is odd.
	 */
	private static final Abstraction newParity() {
		Variable x = new VariableImpl();
		return x.buildApplication(CommonTerms.NOT).buildApplication(
				CommonTerms.K).buildAbstraction(x);
	}

	/**
	 * Creates and returns a Term that represents the successor function for a
	 * numeral under the form (n / 2, n % 2).
	 * <p>
	 * Formula: Succ (C h p) = if p then (C h F) else (C (s h) K)
	 * </p>
	 * 
	 * @return a Term for the successor function of numeral under the form (n /
	 *         2, n % 2).
	 */
	private static final Abstraction newModSucc() {
		Variable x1 = new VariableImpl();
		Variable x2 = new VariableImpl();
		Variable x3 = new VariableImpl();
		Variable c1 = new VariableImpl();
		Variable c2 = new VariableImpl();
		Abstraction p1 = c1.buildApplication(
				CommonTerms.FIRST_IN_COUPLE.buildApplication(x2))
				.buildApplication(CommonTerms.F).buildAbstraction(c1);
		Application s = ChurchNumerals.SUCC
				.buildApplication(CommonTerms.FIRST_IN_COUPLE
						.buildApplication(x3));
		Abstraction p2 = c2.buildApplication(s).buildApplication(CommonTerms.K)
				.buildAbstraction(c2);
		return CommonTerms.SECOND_IN_COUPLE.buildApplication(x1)
				.buildApplication(p1).buildApplication(p2).buildAbstraction(
						new Variable[]{x1, x2, x3});
	}

	/**
	 * Creates and returns a Term that represents a function of a Church numeral
	 * that returns half the value of the argument.
	 * <p>
	 * Formula: Half n = first (n ModSucc ModInt(0))
	 * </p>
	 * 
	 * @return a Term that represents a function of a Church numeral that
	 *         returns half the value of the argument.
	 */
	private static final Abstraction newHalf() {
		Variable xn = new VariableImpl();
		Application pn = xn.buildApplication(MOD_SUCC).buildApplication(
				newModInt(0));
		return CommonTerms.FIRST_IN_COUPLE.buildApplication(pn)
				.buildAbstraction(xn);
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final boolean IS_VERBOSE = false;
	private static final TermReducer TERM_REDUCER = Test.TERM_REDUCER;

	/**
	 * Execute some operations using numerals.
	 * 
	 * @param out
	 *            the PrintStream to print results. May be null.
	 * @param isVerbose
	 *            if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			boolean isOk = true;
			isOk &= test(out, isVerbose, PARITY.buildApplication(ChurchNumerals
					.newInt(0)), true, "Parity 0");
			isOk &= test(out, isVerbose, PARITY.buildApplication(ChurchNumerals
					.newInt(1)), false, "Parity 1");
			isOk &= test(out, isVerbose, PARITY.buildApplication(ChurchNumerals
					.newInt(2)), true, "Parity 2");
			isOk &= test(out, isVerbose, PARITY.buildApplication(ChurchNumerals
					.newInt(3)), false, "Parity 3");
			isOk &= test(out, isVerbose, HALF.buildApplication(ChurchNumerals
					.newInt(0)), 0, "Half 0");
			isOk &= test(out, isVerbose, HALF.buildApplication(ChurchNumerals
					.newInt(1)), 0, "Half 1");
			isOk &= test(out, isVerbose, HALF.buildApplication(ChurchNumerals
					.newInt(2)), 1, "Half 2");
			isOk &= test(out, isVerbose, HALF.buildApplication(ChurchNumerals
					.newInt(3)), 1, "Half 3");
			return isOk;
		} catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static boolean test(PrintStream out, boolean isVerbose, Term term,
			boolean expectedValue, String name) {
		boolean value = TERM_REDUCER.booleanValue(term);
		boolean isOk = value == expectedValue;
		if (isVerbose || !isOk)
			out.println("Modulo2Numerals " + (isOk ? "ok" : "ko") + " " + value
					+ " / " + expectedValue + " (" + name + ")");
		return isOk;
	}

	private static boolean test(PrintStream out, boolean isVerbose, Term term,
			long expectedValue, String name) {
		long value = TERM_REDUCER.numeralValue(term);
		boolean isOk = value == expectedValue;
		if (isVerbose || !isOk)
			out.println("Modulo2Numerals " + (isOk ? "ok" : "ko") + " " + value
					+ " / " + expectedValue + " (" + name + ")");
		return isOk;
	}
}
